"""
Utilities for generating Click options
"""
from collections.abc import MutableSequence, MutableSet, Sequence
from datetime import datetime
from enum import Enum
from functools import update_wrapper
from typing import Any, Callable, Dict, List, Optional, Tuple, Type, Union

import attr
import cattr
import click

from ._compat import get_args, get_origin
from ._core import T, _load_settings, default_loaders
from ._dict_utils import _deep_options, _get_path, _merge_dicts, _set_path
from .attrs import METADATA_KEY, _SecretRepr
from .converters import default_converter, from_dict
from .loaders import Loader


AnyFunc = Callable[..., Any]
Decorator = Callable[[AnyFunc], AnyFunc]
StrDict = Dict[str, Any]


def click_options(
    cls: Type[T],
    loaders: Union[str, List[Loader]],
    converter: Optional[cattr.Converter] = None,
    type_handler: "Optional[TypeHandler]" = None,
) -> Callable[[Callable], Callable]:
    """
    Generate :mod:`click` options for a CLI which override settins loaded via
    :func:`.load_settings()`.

    A single *cls* instance is passed to the decorated function

    Args:
        cls: Attrs class with options (and default values).
        loaders: Either a string with your app name or a list of settings
            :class:`Loader`'s.  If it is a string, use it with
            :func:`~typed_settings.default_loaders()` to get the defalt
            loaders.
        converter: An optional :class:`cattr.Converter` used for converting
            option values to the required type.

            By default, :data:`typed_settings.attrs.converter` is used.
        type_handler: Helps creating proper click options for option types that
            are not natively supported by click.

    Return:
        A decorator for a click command.

    Example:

      .. code-block:: python

         >>> import click
         >>> import typed_settings as ts
         >>>
         >>> @ts.settings
         ... class Settings: ...
         ...
         >>> @click.command()
         ... @ts.click_options(Settings, "example")
         ... def cli(settings):
         ...     print(settings)

    .. versionchanged:: 1.0.0
       Instead of a list of loaders, you can also just pass an application
       name.
    """
    cls = attr.resolve_types(cls)
    options = _deep_options(cls)

    if isinstance(loaders, str):
        loaders = default_loaders(loaders)

    settings_dict = _load_settings(cls, options, loaders)

    converter = converter or default_converter()
    type_handler = type_handler or TypeHandler()

    def pass_settings(f: AnyFunc) -> Decorator:
        """
        Creates a *cls* instances from the settings dict stored in
        :attr:`click.Context.obj` and passes it to the decorated function *f*.
        """

        def new_func(*args, **kwargs):
            ctx = click.get_current_context()
            _merge_dicts(settings_dict, ctx.obj.get("settings"))
            ctx.obj["settings"] = from_dict(settings_dict, cls, converter)
            return f(ctx.obj["settings"], *args, **kwargs)

        return update_wrapper(new_func, f)

    def wrap(f):
        """
        The wrapper that actually decorates a function with all options.
        """
        for oinfo in reversed(options):
            default = _get_default(
                oinfo.field, oinfo.path, settings_dict, converter
            )
            option = _mk_option(
                click.option, oinfo.path, oinfo.field, default, type_handler
            )
            f = option(f)
        f = pass_settings(f)
        return f

    return wrap


def pass_settings(f: AnyFunc) -> AnyFunc:
    """
    Marks a callback as wanting to receive the innermost settings instance as
    first argument.
    """

    def new_func(*args, **kwargs):
        ctx = click.get_current_context()
        node = ctx
        settings = None
        while node is not None:
            if isinstance(node.obj, dict) and "settings" in node.obj:
                settings = node.obj["settings"]
                break
            node = node.parent
        return ctx.invoke(f, settings, *args, **kwargs)

    return update_wrapper(new_func, f)


def handle_datetime(type: type, default: Any) -> StrDict:
    """
    Use :class:`click.DateTime` as option type and convert the default value
    to an ISO string.
    """
    type_info = {
        "type": click.DateTime(
            ["%Y-%m-%d", "%Y-%m-%dT%H:%M:%S", "%Y-%m-%dT%H:%M:%S%z"]
        ),
    }
    if default is not attr.NOTHING:
        type_info["default"] = default.isoformat()
    return type_info


def handle_enum(type: Type[Enum], default: Any) -> StrDict:
    """
    Use :class:`EnumChoice` as option type and use the enum value's name as
    default.
    """
    type_info = {"type": click.Choice(list(type.__members__))}
    if default is not attr.NOTHING:
        # Convert Enum instance to string
        type_info["default"] = default.name

    return type_info


#: Default handlers for click option types.
DEFAULT_TYPES = {
    datetime: handle_datetime,
    Enum: handle_enum,
}


class TypeHandler:
    """
    This class derives type information for Click options from an Attrs field's
    type.

    The class differentitates between specific and generic types (e.g.,
    :samp:`int` vs. :samp:`List[{T}]`.

    Specific types:
        Handlers for specific types can be extended and modified by passing
        a *types* dict to the class.  By default, :data:`DEFAULT_TYPES` is
        used.

        This dict maps Python types to a handler function.  Handler functions
        take the field type and default value and return a dict that is passed
        as keyword arguments to :func:`click.option()`.  This dict should
        contain a ``type`` key and, optionally, an updated ``default``.

        .. code-block:: python

            def handle_mytype(type: type, default: Any) -> Dict[str, Any]:
                type_info = {
                    "type": ClickType(...)
                }
                if default is not attr.NOTHING:
                    type_info["default"] = default.stringify()
                return type_info

        You can use :func:`handle_datetime` and :func:`handle_enum` as
        a sample.

        Types without a handler get no special treatment and cause options to
        look like this: :samp:`click.option(..., type=field_type,
        default=field_default)`.

    Generic types:
        Handlers for generic types cannot be changed.  They either create an
        option with :samp:`multiple=True` or :samp:`nargs={x}`.  Nested types
        are recursively resolved.

        Types that cause :samp:`multiple=True`:

        - :class:`typing.List`
        - :class:`typing.Sequence`
        - :class:`typing.MutableSequence`
        - :class:`typing.Set`
        - :class:`typing.FrozenSet`
        - :class:`typing.MutableSet`

        Types that cause :samp:`nargs={x}`:

        - :class:`typing.Tuple`
        - :class:`typing.NamedTuple`

        Dicts are not (yet) supported.
    """

    def __init__(self, types=None):
        self.types = types or DEFAULT_TYPES
        self.list_types = (
            list,
            Sequence,
            MutableSequence,
            set,
            frozenset,
            MutableSet,
        )
        self.tuple_types = frozenset({tuple})

    def get_type(self, otype: type, default: Any) -> StrDict:
        """
        Analyses the option type and returns updated options.
        """
        origin = get_origin(otype)
        args = get_args(otype)

        if origin is None:
            for target_type, get_type_info in self.types.items():
                if issubclass(otype, target_type):
                    return get_type_info(otype, default)

            return self._handle_basic_types(otype, default)

        else:
            if origin in self.list_types:
                return self._handle_list(otype, default, args)
            elif origin in self.tuple_types:
                return self._handle_tuple(otype, default, args)

            raise TypeError(f"Cannot create click type for: {otype}")

    def _handle_basic_types(self, type: type, default: Any):
        if default is attr.NOTHING:
            type_info = {"type": type}
        else:
            type_info = {"type": type, "default": default}
        return type_info

    def _handle_list(
        self, type: type, default: Any, args: Tuple[Any, ...]
    ) -> StrDict:
        # lists and list-like tuple
        type_info = self.get_type(args[0], attr.NOTHING)
        if default is not attr.NOTHING:
            default = [self.get_type(args[0], d)["default"] for d in default]
            type_info["default"] = default
        type_info["multiple"] = True
        return type_info

    def _handle_tuple(
        self, type: type, default: Any, args: Tuple[Any, ...]
    ) -> StrDict:
        if len(args) == 2 and args[1] == ...:
            return self._handle_list(type, default, args)
        else:
            # "struct" variant of tuple
            if default is attr.NOTHING:
                default = [attr.NOTHING] * len(args)
            dicts = [self.get_type(a, d) for a, d in zip(args, default)]
            type_info = {
                "type": tuple(d["type"] for d in dicts),
                "nargs": len(dicts),
            }
            if all("default" in d for d in dicts):
                type_info["default"] = tuple(d["default"] for d in dicts)
            return type_info


def _get_default(
    field: attr.Attribute,
    path: str,
    settings: StrDict,
    converter: cattr.Converter,
) -> Any:
    """
    Returns the proper default value for an attribute.

    If possible, the default is taken from loaded settings.  Else, use the
    field's default value.
    """
    try:
        # Use loaded settings value
        default = _get_path(settings, path)
    except KeyError:
        # Use field's default
        default = field.default
    else:
        # If the default was found (no KeyError), convert the input value to
        # the proper type.
        # See: https://gitlab.com/sscherfke/typed-settings/-/issues/11
        if field.type:
            default = converter.structure(default, field.type)

    if isinstance(default, attr.Factory):  # type: ignore
        if default.takes_self:
            # There is no instance yet.  Passing ``None`` migh be more correct
            # than passing a fake instance, because it raises an error instead
            # of silently creating a false value. :-?
            default = default.factory(None)
        else:
            default = default.factory()

    return default


def _mk_option(
    option: Callable[..., Decorator],
    path: str,
    field: attr.Attribute,
    default: Any,
    type_handler: TypeHandler,
) -> Decorator:
    """
    Recursively creates click options and returns them as a list.
    """
    opt_name = path.replace(".", "-").replace("_", "-")
    param_decl = f"--{opt_name}"

    def cb(ctx, _param, value):
        if ctx.obj is None:
            ctx.obj = {}
        settings = ctx.obj.setdefault("settings", {})
        _set_path(settings, path, value)
        return value

    metadata = field.metadata.get(METADATA_KEY, {})
    kwargs = {
        "show_default": True,
        "callback": cb,
        "expose_value": False,
        "help": metadata.get("help", ""),
    }

    if isinstance(field.repr, _SecretRepr):
        kwargs["show_default"] = False
        if default is not attr.NOTHING:  # pragma: no cover
            kwargs["help"] = f"{kwargs['help']}  [default: {field.repr('')}]"

    if default is attr.NOTHING:
        kwargs["required"] = True

    if field.type:  # pragma: no cover
        if field.type is bool:
            param_decl = f"{param_decl}/--no-{opt_name}"
        kwargs.update(type_handler.get_type(field.type, default))

    return option(param_decl, **kwargs)



import attr
import pytest

from typed_settings.attrs import SECRET, evolve, option, secret, settings


@settings
class S:
    u: str = option()
    p: str = secret()


class TestAttrExtensions:
    """Tests for attrs extensions."""

    @pytest.fixture
    def inst(self):
        return S(u="spam", p="42")

    def test_secret_str(self, inst):
        assert str(inst) == "S(u='spam', p=***)"

    def test_secret_repr_call(self, inst):
        assert repr(inst) == "S(u='spam', p=***)"

    def test_secret_repr_repr(self):
        assert str(SECRET) == "***"


class TestEvolve:
    """
    Tests for `evolve`.

    Copied from attrs and adjusted/reduced.
    """

    @pytest.fixture(scope="session", name="C")
    def fixture_C(self):
        """
        Return a simple but fully featured attrs class with an x and a y
        attribute.
        """
        import attr

        @attr.s
        class C(object):
            x: str
            y: str

        return C

    def test_validator_failure(self):
        """
        TypeError isn't swallowed when validation fails within evolve.
        """

        @settings
        class C(object):
            a: int = option(validator=attr.validators.instance_of(int))

        with pytest.raises(TypeError) as e:
            evolve(C(a=1), a="some string")
        m = e.value.args[0]

        assert m.startswith("'a' must be <class 'int'>")

    def test_private(self):
        """
        evolve() acts as `__init__` with regards to private attributes.
        """

        @settings
        class C(object):
            _a: str

        assert evolve(C(1), a=2)._a == 2

        with pytest.raises(TypeError):
            evolve(C(1), _a=2)

        with pytest.raises(TypeError):
            evolve(C(1), a=3, _a=2)

    def test_non_init_attrs(self):
        """
        evolve() handles `init=False` attributes.
        """

        @settings
        class C(object):
            a: str
            b: int = option(init=False, default=0)

        assert evolve(C(1), a=2).a == 2

    def test_regression_attrs_classes(self):
        """
        evolve() can evolve fields that are instances of attrs classes.

        Regression test for #804
        """

        @settings
        class Child(object):
            param2: str

        @settings
        class Parent(object):
            param1: Child

        obj2a = Child(param2="a")
        obj2b = Child(param2="b")

        obj1a = Parent(param1=obj2a)

        assert Parent(param1=Child(param2="b")) == evolve(obj1a, param1=obj2b)

    def test_recursive(self):
        """
        evolve() recursively evolves nested attrs classes when a dict is
        passed for an attribute.
        """

        @settings
        class N2(object):
            e: int

        @settings
        class N1(object):
            c: N2
            d: int

        @settings
        class C(object):
            a: N1
            b: int

        c1 = C(N1(N2(1), 2), 3)
        c2 = evolve(c1, a={"c": {"e": 23}}, b=42)

        assert c2 == C(N1(N2(23), 2), 42)

    def test_recursive_attrs_classes(self):
        """
        evolve() can evolve fields that are instances of attrs classes.
        """

        @settings
        class Child:
            param2: str

        @settings
        class Parent:
            param1: Child

        obj2a = Child(param2="a")
        obj2b = Child(param2="b")

        obj1a = Parent(param1=obj2a)

        result = evolve(obj1a, param1=obj2b)
        assert result.param1 is obj2b



import pytest

from typed_settings import _file_utils as fu


@pytest.mark.parametrize(
    "args, start, expected",
    [
        # File found
        pytest.param(["s.toml"], ".", "s.toml", id="found-from-pwd"),
        pytest.param(["s.toml"], "src", "s.toml", id="found-from-subdir"),
        pytest.param(["s.toml"], "src/a/x", "s.toml", id="found-from-subdirs"),
        pytest.param(
            ["s.toml", "."],
            "src/a/x",
            "s.toml",
            id="found-from-subdirs-stop-project",
        ),
        pytest.param(
            ["s.toml", fu.ROOT_DIR, ["setup.py"]],
            "src/a/x",
            "s.toml",
            id="found-from-subdirs-stop-existing-file",
        ),
        pytest.param(
            ["s.toml", fu.ROOT_DIR, ["spam"]],
            "src/a/x",
            "s.toml",
            id="found-from-subdirs-found-stop-nonexisting-file",
        ),
        pytest.param(
            ["s.toml", "x"],
            "src/a/x",
            "s.toml",
            id="found-from-subdirs-stop-nonexisting-rootdir",
        ),
        pytest.param(
            ["s.toml", "x", ["spam"]],
            "src/a/x",
            "s.toml",
            id="found-from-subdirs-stop-nonexisting-rootdir-and-file",
        ),
        # File not found
        pytest.param(
            ["s.toml", "src/a"],
            "src/a/x",
            "src/a/x/s.toml",
            id="not-found-stop-at-subdir",
        ),
        pytest.param(
            ["s.toml", fu.ROOT_DIR, ["stop"]],
            "src/a/x",
            "src/a/x/s.toml",
            id="not-found-stop-at-stop-file",
        ),
        pytest.param(["spam"], ".", "spam", id="not-found-not-exists"),
        pytest.param(
            ["spam"],
            "src",
            "src/spam",
            id="not-found-not-exists-start-from-subdir",
        ),
        pytest.param(
            ["spam", "xxx", []],
            ".",
            "spam",
            id="not-found-non-existing-rootdir-and-stopfile",
        ),
    ],
)
def test_find(args, start, expected, tmp_path, monkeypatch):
    """find() always returns a path, never raises something."""
    for p in [".git", "src/a/x", "src/a/y"]:
        tmp_path.joinpath(p).mkdir(parents=True, exist_ok=True)
    for p in ["setup.py", "s.toml", "src/stop"]:
        tmp_path.joinpath(p).touch()

    monkeypatch.chdir(tmp_path.joinpath(start))
    if len(args) > 1:
        if isinstance(args[1], str):
            args[1] = tmp_path.joinpath(args[1])
    result = fu.find(*args)
    assert result == tmp_path.joinpath(expected)



"""
Core functionality for loading settings.
"""
import logging
from pathlib import Path
from typing import Any, Dict, Iterable, List, Type, Union

import attr
import cattr

from ._dict_utils import _deep_options, _merge_dicts, _set_path
from .attrs import METADATA_KEY
from .converters import default_converter, from_dict
from .loaders import EnvLoader, FileLoader, Loader, TomlFormat
from .types import AUTO, OptionList, T, _Auto


LOGGER = logging.getLogger(METADATA_KEY)


def default_loaders(
    appname: str,
    config_files: Iterable[Union[str, Path]] = (),
    *,
    config_file_section: Union[str, _Auto] = AUTO,
    config_files_var: Union[None, str, _Auto] = AUTO,
    env_prefix: Union[None, str, _Auto] = AUTO,
) -> List[Loader]:
    """
    Return a list of default settings loaders that are used by :func:`load()`.

    These loaders are:

    #. A :class:`.FileLoader` loader configured with the :class:`TomlFormat`
    #. An :class:`.EnvLoader`

    The :class:`.FileLoader` will load files from *config_files* and from the
    environment variable *config_files_var*.

    Args:
        appname: Your application's name – used to derive defaults for the
          remaining args.

        config_files: Load settings from these files.  The last one has the
          highest precedence.

        config_file_section: Name of your app's section in the config file.
          By default, use *appname* (in lower case and with "_" replaced by
          "-".

        config_files_var: Load list of settings files from this environment
          variable.  By default, use :code:`{APPNAME}_SETTINGS`.  Multiple
          paths have to be separated by ":".  The last file has the highest
          precedence.  All files listed in this var have higher precedence than
          files from *config_files*.

          Set to ``None`` to disable this feature.

        env_prefix: Load settings from environment variables with this prefix.
          By default, use *APPNAME_*.

          Set to ``None`` to disable loading env vars.

    Return:
        A list of :class:`.Loader` instances.
    """
    loaders: List[Loader] = []

    section = (
        appname.lower().replace("_", "-")
        if isinstance(config_file_section, _Auto)
        else config_file_section
    )
    var_name = (
        f"{appname.upper()}_SETTINGS".replace("-", "_")
        if isinstance(config_files_var, _Auto)
        else config_files_var
    )
    loaders.append(
        FileLoader(
            files=config_files,
            env_var=var_name,
            formats={"*.toml": TomlFormat(section)},
        )
    )

    if env_prefix is None:
        LOGGER.debug("Loading settings from env vars is disabled.")
    else:
        prefix = (
            f"{appname.upper()}_"
            if isinstance(env_prefix, _Auto)
            else env_prefix
        )
        loaders.append(EnvLoader(prefix=prefix))

    return loaders


def load(
    cls: Type[T],
    appname: str,
    config_files: Iterable[Union[str, Path]] = (),
    *,
    config_file_section: Union[str, _Auto] = AUTO,
    config_files_var: Union[None, str, _Auto] = AUTO,
    env_prefix: Union[None, str, _Auto] = AUTO,
) -> T:
    """
    Load settings for *appname* and return an instance of *cls*

    This function is a shortcut for :func:`load_settings()` with
    :func:`default_loaders()`.

    Settings are loaded from *config_files* and from the files specified
    via the *config_files_var* environment variable.  Settings can also be
    overridden via environment variables named like the corresponding setting
    and prefixed with *env_prefix*.

    Settings precedence (from lowest to highest priority):

    - Default value from *cls*
    - First file from *config_files*
    - ...
    - Last file from *config_files*
    - First file from *config_files_var*
    - ...
    - Last file from *config_files_var*
    - Environment variable :code:`{env_prefix}_{SETTING}`

    Config files (both, explicitly specified, and loaded from an environment
    variable) are optional by default.  You can prepend an ``!`` to their path
    to mark them as mandatory (e.g., `!/etc/credentials.toml`).  An error is
    raised if a mandatory file does not exist.

    Args:
        cls: Attrs class with default settings.

        appname: Your application's name.  Used to derive defaults for the
          remaining args.

        config_files: Load settings from these files.

        config_file_section: Name of your app's section in the config file.
          By default, use *appname* (in lower case and with "_" replaced by
          "-".

        config_files_var: Load list of settings files from this environment
          variable.  By default, use :code:`{APPNAME}_SETTINGS`.  Multiple
          paths have to be separated by ":".  The last file has the highest
          precedence.  All files listed in this var have higher precedence than
          files from *config_files*.

          Set to ``None`` to disable this feature.

        env_prefix: Load settings from environment variables with this prefix.
          By default, use *APPNAME_*.

          Set to ``None`` to disable loading env vars.

    Return:
        An instance of *cls* populated with settings from settings files and
        environment variables.

    Raise:
        UnknownFormatError: When no :class:`FileFormat` is configured for a
            loaded file.
        ConfigFileNotFoundError: If *path* does not exist.
        ConfigFileLoadError: If *path* cannot be read/loaded/decoded.
        InvalidOptionsError: If invalid settings have been found.
        InvalidValueError: If a value cannot be converted to the correct type.
    """
    loaders = default_loaders(
        appname=appname,
        config_files=config_files,
        config_file_section=config_file_section,
        config_files_var=config_files_var,
        env_prefix=env_prefix,
    )
    settings = _load_settings(
        cls=cls,
        options=_deep_options(cls),
        loaders=loaders,
    )

    converter = default_converter()
    return from_dict(settings, cls, converter)


def load_settings(
    cls: Type[T],
    loaders: List[Loader],
    converter: cattr.Converter = None,
) -> T:
    """
    Load settings defined by the class *cls* and return an instance of it.

    Args:
        cls: Attrs class with options (and default values).
        loaders: A list of settings :class:`Loader`'s.
        converter: An optional :class:`cattr.Converter` used for converting
            option values to the required type.

            By default, :data:`typed_settings.attrs.converter` is used.

    Return:
        An instance of *cls* populated with settings from the defined loaders.

    Raise:
        TsError: Depending on the configured loaders, any subclass of this
            exception.
    """
    if converter is None:
        converter = default_converter()
    settings = _load_settings(
        cls=cls,
        options=_deep_options(cls),
        loaders=loaders,
    )
    return from_dict(settings, cls, converter)


def _load_settings(
    cls: type,
    options: OptionList,
    loaders: List[Loader],
) -> Dict[str, Any]:
    """
    Loads settings for *options* and returns them as dict.

    This function makes it easier to extend settings since it returns a dict
    that can easily be updated.
    """
    settings: Dict[str, Any] = {}

    # Populate dict with default settings.  This avoids problems with nested
    # settings classes for which no settings are loaded.
    for opt in options:
        if opt.field.default is attr.NOTHING:
            continue
        if isinstance(opt.field.default, attr.Factory):  # type: ignore
            continue
        _set_path(settings, opt.path, opt.field.default)

    loaded_settings = [loader(cls, options) for loader in loaders]

    for ls in loaded_settings:
        _merge_dicts(settings, ls)

    return settings



"""
Test that all public functions are properly exposed.
"""
import pytest

import typed_settings as ts


@ts.settings
class Settings:
    u: str = ts.option()
    p: str = ts.secret()


@ts.settings(frozen=True)
class FrozenSettings:
    u: str = ts.option()
    p: str = ts.secret()


classes = [Settings, FrozenSettings]


@pytest.mark.parametrize("cls", classes)
def test_load(cls, tmp_path):
    """We can load settings with a class decorated with our decorator."""
    f = tmp_path.joinpath("cfg.toml")
    f.write_text('[test]\nu = "spam"\np = "eggs"\n')
    settings = ts.load(cls, "test", [f])
    assert settings == cls("spam", "eggs")


@pytest.mark.parametrize("cls", classes)
def test_load_settings(cls, tmp_path):
    """We can load settings with a class decorated with our decorator."""
    f = tmp_path.joinpath("cfg.toml")
    f.write_text('[test]\nu = "spam"\np = "eggs"\n')
    settings = ts.load(cls, "test", [f])
    assert settings == cls("spam", "eggs")


def test_dir():
    names = dir(ts)
    assert names == [
        "EnvLoader",
        "FileLoader",
        "TomlFormat",
        "click_options",
        "default_converter",
        "default_loaders",
        "evolve",
        "find",
        "load",
        "load_settings",
        "option",
        "pass_settings",
        "register_strlist_hook",
        "secret",
        "settings",
    ]



"""
Typed settings
"""
from typing import Any, List

from ._core import default_loaders, load, load_settings
from ._file_utils import find
from .attrs import evolve, option, secret, settings
from .converters import default_converter, register_strlist_hook
from .loaders import EnvLoader, FileLoader, TomlFormat


__all__ = [
    "EnvLoader",
    "FileLoader",
    "TomlFormat",
    "click_options",
    "default_converter",
    "default_loaders",
    "evolve",
    "find",
    "load",
    "load_settings",
    "option",
    "pass_settings",
    "register_strlist_hook",
    "secret",
    "settings",
]


def __getattr__(name: str) -> Any:
    if name == "click_options":
        from .click_utils import click_options

        return click_options
    if name == "pass_settings":
        from .click_utils import pass_settings

        return pass_settings

    raise AttributeError(f"module '{__name__}' has no attribute '{name}'")


def __dir__() -> List[str]:
    return __all__



"""
Additional attrs validators

.. deprecated:: 1.0.1
   Use :mod:`attrs.validators` instead.  Will be removed in 2.0.0.
"""
from attrs.validators import ge, gt, le, lt
from attrs.validators import max_len
from attrs.validators import max_len as maxlen


__all__ = [
    "lt",
    "le",
    "ge",
    "gt",
    "max_len",
    "maxlen",
]



from enum import Enum

import click

import typed_settings as ts


class PyVersion(Enum):
    py37 = "3.7"
    py38 = "3.8"
    py39 = "3.9"


@ts.settings
class Settings:
    line_length: int = 88
    skip_string_normalization: bool = False
    # target_version: Optional[PyVersion] = None  # Does not yet work
    target_version: PyVersion = PyVersion.py37


@click.command()
@ts.click_options(
    Settings,
    ts.default_loaders(
        appname="black",
        config_files=[ts.find("pyproject.toml")],
        config_file_section="tool.black",
        env_prefix=None,
    ),
)
def cli(settings):
    print(settings)


if __name__ == "__main__":
    cli()



from datetime import datetime, timezone
from enum import Enum
from pathlib import Path
from typing import (
    Any,
    Callable,
    Dict,
    FrozenSet,
    Generic,
    List,
    MutableSequence,
    MutableSet,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    TypeVar,
)

import attr
import click
import click.testing
import pytest
from _pytest.python import Metafunc

from typed_settings import (
    click_options,
    click_utils,
    default_converter,
    default_loaders,
    option,
    pass_settings,
    secret,
    settings,
)


T = TypeVar("T")


class CliResult(click.testing.Result, Generic[T]):
    settings: Optional[T]


Cli = Callable[..., CliResult[T]]


def make_cli(settings_cls: Type[T]) -> Cli:
    """
    Return a function that invokes a Click CLI with Typed Settings options
    for *settings_cls*.

    That functions returns a :class:`CliResult` with the loaded settings
    instance (:attr:`CliResult.settings`).
    """

    class Runner(click.testing.CliRunner):
        settings: Optional[T]

        def invoke(self, *args, **kwargs) -> CliResult:
            result = super().invoke(*args, **kwargs)
            cli_result: CliResult[T] = CliResult(**result.__dict__)
            try:
                cli_result.settings = self.settings
            except AttributeError:
                cli_result.settings = None
            return cli_result

    runner = Runner()

    def run(*args, **kwargs) -> CliResult:
        @click.group(invoke_without_command=True)
        @click_options(settings_cls, default_loaders("test"))
        def cli(settings: T) -> None:
            runner.settings = settings

        return runner.invoke(cli, args, **kwargs)

    return run


def test_simple_cli():
    """
    "click_options()" uses the default loaders if you just pass an app name.
    """

    @settings
    class Settings:
        o: int

    loaded = []

    @click.command()
    @click_options(Settings, "test")
    def cli(settings: Settings) -> None:
        loaded.append(settings)

    result = click.testing.CliRunner().invoke(cli, ["--o=3"])
    assert result.exit_code == 0
    assert loaded == [Settings(3)]


@pytest.mark.parametrize(
    "default, path, type, settings, expected",
    [
        (attr.NOTHING, "a", int, {"a": 3}, 3),
        (attr.NOTHING, "a", int, {}, attr.NOTHING),
        (2, "a", int, {}, 2),
        (attr.Factory(list), "a", List[int], {}, []),
    ],
)
def test_get_default(default, path, type, settings, expected):
    converter = default_converter()
    field = attr.Attribute(
        "test", default, None, None, None, None, None, None, type=type
    )
    result = click_utils._get_default(field, path, settings, converter)
    assert result == expected


def test_get_default_factory():
    """
    If the factory "takes self", ``None`` is passed since we do not yet have
    an instance.
    """

    def factory(self) -> str:
        assert self is None
        return "eggs"

    default = attr.Factory(factory, takes_self=True)
    field = attr.Attribute("test", default, None, None, None, None, None, None)
    result = click_utils._get_default(field, "a", {}, default_converter())
    assert result == "eggs"


def test_no_default(monkeypatch):
    """
    cli_options without a default are mandatory/required.
    """

    @settings
    class Settings:
        a: str
        b: str

    monkeypatch.setenv("TEST_A", "spam")  # This makes only "S.b" mandatory!

    @click.command()
    @click_options(Settings, default_loaders("test"))
    def cli(settings):
        pass

    runner = click.testing.CliRunner()
    result = runner.invoke(cli, [])
    assert result.output == (
        "Usage: cli [OPTIONS]\n"
        "Try 'cli --help' for help.\n"
        "\n"
        "Error: Missing option '--b'.\n"
    )
    assert result.exit_code == 2


def test_help_text():
    """
    cli_options/secrets can specify a help text for click cli_options.
    """

    @settings
    class Settings:
        a: str = option(default="spam", help="Help for 'a'")
        b: str = secret(default="eggs", help="bbb")

    @click.command()
    @click_options(Settings, default_loaders("test"))
    def cli(settings):
        pass

    runner = click.testing.CliRunner()
    result = runner.invoke(cli, ["--help"])
    assert result.output == (
        "Usage: cli [OPTIONS]\n"
        "\n"
        "Options:\n"
        "  --a TEXT  Help for 'a'  [default: spam]\n"
        "  --b TEXT  bbb  [default: ***]\n"
        "  --help    Show this message and exit.\n"
    )
    assert result.exit_code == 0


def test_long_name():
    """
    Underscores in option names are replaces with "-" in Click cli_options.
    """

    @settings
    class Settings:
        long_name: str = "val"

    @click.command()
    @click_options(Settings, default_loaders("test"))
    def cli(settings):
        pass

    runner = click.testing.CliRunner()
    result = runner.invoke(cli, ["--help"])
    assert result.output == (
        "Usage: cli [OPTIONS]\n"
        "\n"
        "Options:\n"
        "  --long-name TEXT  [default: val]\n"
        "  --help            Show this message and exit.\n"
    )
    assert result.exit_code == 0


def test_click_default_from_settings(monkeypatch, tmp_path):
    """
    If a setting is set in a config file, that value is being used as default
    for click cli_options - *not* the default defined in the Settings class.
    """

    tmp_path.joinpath("settings.toml").write_text('[test]\na = "x"\n')
    spath = tmp_path.joinpath("settings2.toml")
    spath.write_text('[test]\nb = "y"\n')
    monkeypatch.setenv("TEST_SETTINGS", str(spath))
    monkeypatch.setenv("TEST_C", "z")

    @settings
    class Settings:
        a: str
        b: str
        c: str
        d: str

    @click.command()
    @click_options(
        Settings, default_loaders("test", [tmp_path.joinpath("settings.toml")])
    )
    def cli(settings):
        print(settings)

    runner = click.testing.CliRunner()
    result = runner.invoke(cli, ["--help"])
    assert result.output == (
        "Usage: cli [OPTIONS]\n"
        "\n"
        "Options:\n"
        "  --a TEXT  [default: x]\n"
        "  --b TEXT  [default: y]\n"
        "  --c TEXT  [default: z]\n"
        "  --d TEXT  [required]\n"
        "  --help    Show this message and exit.\n"
    )
    assert result.exit_code == 0


def test_unsupported_generic():
    @settings
    class Settings:
        opt: Dict[int, int]

    with pytest.raises(TypeError, match="Cannot create click type"):

        @click.command()
        @click_options(Settings, default_loaders("test"))
        def cli(settings):
            pass


class LeEnum(Enum):
    spam = "Le spam"
    eggs = "Le eggs"


class TestClickParamTypes:
    """
    Test the behavior of click options for various parameter types.

    """

    class ClickParamBase:
        """
        Base class for test parameters.

        Sublcasses must define:

        - A settings class
        - A list of expected "--help" outputs for each option
        - Optionally, required arguments for options with no default.
        - An instance of the settings class with expected default option
          values.
        - A list of Cli options and the expected settings result.
        """

        @settings
        class Settings:
            pass

        expected_help: List[str] = []

        env_vars: Dict[str, str] = {}
        expected_env_var_defaults: List[str] = []

        default_options: List[str] = []
        expected_defaults: Any = None

        cli_options: List[str] = []
        expected_settings: Any = None

        _classes: List["TestClickParamTypes.ClickParamBase"] = []

        def __init_subclass__(cls, **kwargs):
            super().__init_subclass__(**kwargs)
            cls._classes.append(cls)

    class ClickBoolParam(ClickParamBase):
        """
        Test boolean flags.
        """

        @settings
        class Settings:
            a: bool
            b: bool = True
            c: bool = False

        expected_help = [
            "  --a / --no-a  [required]",
            "  --b / --no-b  [default: b]",
            "  --c / --no-c  [default: no-c]",
        ]

        env_vars = {"A": "1", "B": "0"}
        expected_env_var_defaults = [
            "  --a / --no-a  [default: a]",
            "  --b / --no-b  [default: no-b]",
            "  --c / --no-c  [default: no-c]",
        ]

        default_options = ["--a"]
        expected_defaults = Settings(True, True, False)

        cli_options = ["--no-a", "--no-b", "--c"]
        expected_settings = Settings(False, False, True)

    class IntFloatStrParam(ClickParamBase):
        """
        Test int, float and str cli_options.
        """

        @settings
        class Settings:
            a: str = option(default="spam")
            b: str = secret(default="spam")
            c: int = 0
            d: float = 0

        expected_help = [
            "  --a TEXT     [default: spam]",
            "  --b TEXT     [default: ***]",
            "  --c INTEGER  [default: 0]",
            "  --d FLOAT    [default: 0.0]",
        ]

        env_vars = {"A": "eggs", "B": "bacon", "C": "42", "D": "3.14"}
        expected_env_var_defaults = [
            "  --a TEXT     [default: eggs]",
            "  --b TEXT     [default: ***]",
            "  --c INTEGER  [default: 42]",
            "  --d FLOAT    [default: 3.14]",
        ]

        expected_defaults = Settings()

        cli_options = ["--a=eggs", "--b=pwd", "--c=3", "--d=3.1"]
        expected_settings = Settings(a="eggs", b="pwd", c=3, d=3.1)

    class DateTimeParam(ClickParamBase):
        """
        Test datetime cli_options.
        """

        @settings
        class Settings:
            a: datetime = datetime.fromtimestamp(0, timezone.utc)
            b: datetime = datetime.fromtimestamp(0, timezone.utc)
            c: datetime = datetime.fromtimestamp(0, timezone.utc)

        expected_help = [
            "  --a [%Y-%m-%d|%Y-%m-%dT%H:%M:%S|%Y-%m-%dT%H:%M:%S%z]",
            "                                  [default: 1970-01-01T00:00:00+00:00]",  # noqa: E501
            "  --b [%Y-%m-%d|%Y-%m-%dT%H:%M:%S|%Y-%m-%dT%H:%M:%S%z]",
            "                                  [default: 1970-01-01T00:00:00+00:00]",  # noqa: E501
            "  --c [%Y-%m-%d|%Y-%m-%dT%H:%M:%S|%Y-%m-%dT%H:%M:%S%z]",
            "                                  [default: 1970-01-01T00:00:00+00:00]",  # noqa: E501
        ]

        env_vars = {
            "A": "2021-05-04T13:37:00Z",
            "B": "2021-05-04T13:37:00",
            "C": "2021-05-04",
        }
        expected_env_var_defaults = [
            "  --a [%Y-%m-%d|%Y-%m-%dT%H:%M:%S|%Y-%m-%dT%H:%M:%S%z]",
            "                                  [default: 2021-05-04T13:37:00+00:00]",  # noqa: E501
            "  --b [%Y-%m-%d|%Y-%m-%dT%H:%M:%S|%Y-%m-%dT%H:%M:%S%z]",
            "                                  [default: 2021-05-04T13:37:00]",  # noqa: E501
            "  --c [%Y-%m-%d|%Y-%m-%dT%H:%M:%S|%Y-%m-%dT%H:%M:%S%z]",
            "                                  [default: 2021-05-04T00:00:00]",  # noqa: E501
        ]

        expected_defaults = Settings()

        cli_options = [
            "--a=2020-05-04",
            "--b=2020-05-04T13:37:00",
            "--c=2020-05-04T13:37:00+00:00",
        ]
        expected_settings = Settings(
            datetime(2020, 5, 4),
            datetime(2020, 5, 4, 13, 37),
            datetime(2020, 5, 4, 13, 37, tzinfo=timezone.utc),
        )

    class EnumParam(ClickParamBase):
        """
        Test enum cli_options
        """

        @settings
        class Settings:
            a: LeEnum
            b: LeEnum = LeEnum.spam

        expected_help = [
            "  --a [spam|eggs]  [required]",
            "  --b [spam|eggs]  [default: spam]",
        ]

        env_vars = {"A": "spam", "B": "eggs"}
        expected_env_var_defaults = [
            "  --a [spam|eggs]  [default: spam]",
            "  --b [spam|eggs]  [default: eggs]",
        ]

        default_options = ["--a=spam"]
        expected_defaults = Settings(a=LeEnum.spam)

        cli_options = ["--a=spam", "--b=eggs"]
        expected_settings = Settings(LeEnum.spam, LeEnum.eggs)

    class PathParam(ClickParamBase):
        """
        Test Path cli_options
        """

        @settings
        class Settings:
            a: Path = Path("/")

        expected_help = ["  --a PATH  [default: /]"]

        env_vars = {"A": "/spam/eggs"}
        expected_env_var_defaults = ["  --a PATH  [default: /spam/eggs]"]

        expected_defaults = Settings()

        cli_options = ["--a=/spam"]
        expected_settings = Settings(Path("/spam"))

    class NestedParam(ClickParamBase):
        """
        Test cli_options for nested settings
        """

        @settings
        class Settings:
            @settings
            class Nested:
                a: str = "nested"
                b: int = 0

            n: Nested = Nested()

        expected_help = [
            "  --n-a TEXT     [default: nested]",
            "  --n-b INTEGER  [default: 0]",
        ]

        env_vars = {"N_A": "spam", "N_B": "42"}
        expected_env_var_defaults = [
            "  --n-a TEXT     [default: spam]",
            "  --n-b INTEGER  [default: 42]",
        ]

        expected_defaults = Settings()

        cli_options = ["--n-a=eggs", "--n-b=3"]
        expected_settings = Settings(Settings.Nested("eggs", 3))

    class ListParam(ClickParamBase):
        """
        Lists (and friends) use "multiple=True".
        """

        @settings
        class Settings:
            a: List[int]
            b: Sequence[datetime] = [datetime(2020, 5, 4)]
            c: MutableSequence[int] = []
            d: Set[int] = set()
            e: MutableSet[int] = set()
            f: FrozenSet[int] = frozenset()

        expected_help = [
            "  --a INTEGER                     [required]",
            "  --b [%Y-%m-%d|%Y-%m-%dT%H:%M:%S|%Y-%m-%dT%H:%M:%S%z]",
            "                                  [default: 2020-05-04T00:00:00]",
            "  --c INTEGER",
            "  --d INTEGER",
            "  --e INTEGER",
            "  --f INTEGER",
        ]

        env_vars = {
            "A": "1:2",
            "B": "2021-01-01:2021-01-02",  # Dates with times wont work here!
        }
        expected_env_var_defaults = [
            "  --a INTEGER                     [default: 1, 2]",
            "  --b [%Y-%m-%d|%Y-%m-%dT%H:%M:%S|%Y-%m-%dT%H:%M:%S%z]",
            "                                  [default: 2021-01-01T00:00:00,",
            "                                  2021-01-02T00:00:00]",
            "  --c INTEGER",
            "  --d INTEGER",
            "  --e INTEGER",
            "  --f INTEGER",
        ]

        default_options = ["--a=1"]
        expected_defaults = Settings(a=[1])

        cli_options = [
            "--a=1",
            "--a=2",
            "--b=2020-01-01",
            "--b=2020-01-02",
            "--c=3",
            "--d=4",
            "--e=5",
            "--f=6",
        ]
        expected_settings = Settings(
            [1, 2],
            [datetime(2020, 1, 1), datetime(2020, 1, 2)],
            [3],
            {4},
            {5},
            frozenset({6}),
        )

    class TupleParam(ClickParamBase):
        """
        Tuples are handled either like the list variant with multiple=True or
        like the struct variant with nargs=x.
        """

        @settings
        class Settings:
            a: Tuple[int, ...] = (0,)
            b: Tuple[int, float, str] = (0, 0.0, "")

        expected_help = [
            "  --a INTEGER                  [default: 0]",
            "  --b <INTEGER FLOAT TEXT>...  [default: 0, 0.0, ]",
        ]

        env_vars = {"A": "1:2", "B": "42:3.14:spam"}
        expected_env_var_defaults = [
            "  --a INTEGER                  [default: 1, 2]",
            "  --b <INTEGER FLOAT TEXT>...  [default: 42, 3.14, spam]",
        ]

        expected_defaults = Settings()

        cli_options = ["--a=1", "--a=2", "--b", "1", "2.3", "spam"]
        expected_settings = Settings((1, 2), (1, 2.3, "spam"))

    class NestedTupleParam(ClickParamBase):
        """
        Lists of tuples use "multiple=True" and "nargs=x".
        """

        @settings
        class Settings:
            a: List[Tuple[int, int]] = option(factory=list)

        expected_help = [
            "  --a <INTEGER INTEGER>...",
        ]

        # A list of tuples cannot be loaded with the default converter,
        # so we skip this test here.
        env_vars: Dict[str, Any] = {}
        expected_env_var_defaults = [
            "  --a <INTEGER INTEGER>...",
        ]

        expected_defaults = Settings()

        cli_options = ["--a", "1", "2", "--a", "3", "4"]
        expected_settings = Settings([(1, 2), (3, 4)])

    def pytest_generate_tests(self, metafunc: Metafunc) -> None:
        params = []
        fixtures = ["cli"] + [
            n for n in metafunc.fixturenames if hasattr(self.ClickParamBase, n)
        ]
        for param_cls in self.ClickParamBase._classes:
            argvals = []
            for name in fixtures:
                if name == "cli":
                    argvals.append(make_cli(param_cls.Settings))
                else:
                    argvals.append(getattr(param_cls, name))
            params.append(pytest.param(*argvals, id=param_cls.__name__))  # type: ignore  # noqa: E501

        metafunc.parametrize(fixtures, params)

    def test_help(self, cli: Cli[T], expected_help: List[str]):
        """
        The genereated CLI has a proper help output.
        """
        result = cli("--help")

        # fmt: off
        assert result.output.splitlines()[:-1] == [
            "Usage: cli [OPTIONS] COMMAND [ARGS]...",
            "",
            "Options:",
        ] + expected_help
        assert result.exit_code == 0
        # fmt: on

    def test_defaults_from_envvars(
        self,
        cli: Cli[T],
        env_vars: Dict[str, str],
        expected_env_var_defaults: List[str],
        monkeypatch: pytest.MonkeyPatch,
    ):
        """
        Previously loaded settings (e.g., from env vars) can be converted
        to click options.

        Regression test for #11
        """
        for var, val in env_vars.items():
            monkeypatch.setenv(f"TEST_{var}", val)

        result = cli("--help")

        # fmt: off
        assert result.output.splitlines()[:-1] == [
            "Usage: cli [OPTIONS] COMMAND [ARGS]...",
            "",
            "Options:",
        ] + expected_env_var_defaults
        assert result.exit_code == 0
        # fmt: on

    def test_defaults(
        self, cli: Cli[T], default_options: List[str], expected_defaults: T
    ):
        """
        Arguments of the generated CLI have default values.
        """
        result = cli(*default_options)
        assert result.output == ""
        assert result.exit_code == 0
        assert result.settings == expected_defaults

    def test_options(
        self, cli: Cli[T], cli_options: List[str], expected_settings: T
    ):
        """
        Default values can be overriden by passing the corresponding args.
        """
        result = cli(*cli_options)
        assert result.output == ""
        assert result.exit_code == 0
        assert result.settings == expected_settings


class TestPassSettings:
    """Tests for pass_settings()."""

    @settings
    class Settings:
        opt: str = ""

    def test_pass_settings(self):
        """
        A subcommand can receive the settings via the `pass_settings`
        decorator.
        """

        @click.group()
        @click_options(self.Settings, default_loaders("test"))
        def cli(settings):
            pass

        @cli.command()
        @pass_settings
        def cmd(settings):
            print(settings)
            assert settings == self.Settings(opt="spam")

        runner = click.testing.CliRunner()
        result = runner.invoke(cli, ["--opt=spam", "cmd"])
        assert result.output == "TestPassSettings.Settings(opt='spam')\n"
        assert result.exit_code == 0

    def test_pass_settings_no_settings(self):
        """
        Pass ``None`` if no settings are defined.
        """

        @click.group()
        def cli():
            pass

        @cli.command()
        @pass_settings
        def cmd(settings):
            print(settings)
            assert settings is None

        runner = click.testing.CliRunner()
        result = runner.invoke(cli, ["cmd"])
        assert result.output == "None\n"
        assert result.exit_code == 0

    def test_pass_in_parent_context(self):
        """
        The decorator can be used in the same context as "click_options()".
        This makes no sense, but works.
        """

        @click.command()
        @click_options(self.Settings, default_loaders("test"))
        @pass_settings
        def cli(s1, s2):
            click.echo(s1 == s2)

        runner = click.testing.CliRunner()
        result = runner.invoke(cli, ["--opt=spam"])
        assert result.output == "True\n"
        assert result.exit_code == 0



from typing import Any

from attr import fields, has, resolve_types

from .types import OptionInfo, OptionList, SettingsDict


def _deep_options(cls: type) -> OptionList:
    """
    Recursively iterates *cls* and nested attrs classes and returns a flat
    list of *(path, Attribute, type)* tuples.

    Args:
        cls: The class whose attributes will be listed.

    Returns:
        The flat list of attributes of *cls* and possibly nested attrs classes.
        *path* is a dot (``.``) separted path to the attribute, e.g.
        ``"parent_attr.child_attr.grand_child_attr``.

    Raises:
        NameError: if the type annotations can not be resolved.  This is, e.g.,
          the case when recursive classes are being used.
    """
    cls = resolve_types(cls)
    result = []

    def iter_attribs(r_cls: type, prefix: str) -> None:
        for field in fields(r_cls):
            if field.type is not None and has(field.type):
                iter_attribs(field.type, f"{prefix}{field.name}.")
            else:
                result.append(
                    OptionInfo(f"{prefix}{field.name}", field, r_cls)
                )

    iter_attribs(cls, "")
    return result


def _get_path(dct: SettingsDict, path: str) -> Any:
    """
    Performs a nested dict lookup for *path* and returns the result.

    Calling ``_get_path(dct, "a.b")`` is equivalent to ``dict["a"]["b"]``.

    Args:
        dct: The source dict
        path: The path to look up.  It consists of the dot-separated nested
          keys.

    Returns:
        The looked up value.

    Raises:
        KeyError: if a key in *path* does not exist.
    """
    for part in path.split("."):
        dct = dct[part]
    return dct


def _set_path(dct: SettingsDict, path: str, val: Any) -> None:
    """
    Sets a value to a nested dict and automatically creates missing dicts
    should they not exist.

    Calling ``_set_path(dct, "a.b", 3)`` is equivalent to ``dict["a"]["b"]
    = 3``.

    Args:
        dct: The dict that should contain the value
        path: The (nested) path, a dot-separated concatenation of keys.
        val: The value to set
    """
    *parts, key = path.split(".")
    for part in parts:
        dct = dct.setdefault(part, {})
    dct[key] = val


def _merge_dicts(d1: SettingsDict, d2: SettingsDict) -> None:
    """
    Recursively merges *d2* into *d1*.  *d1* is modified in place.

    Args:
        d1: The base dict that will be modified.
        d2: The dict that will be merged into d1, remains unchanged.

    """
    for k, v in d2.items():
        if k in d1 and isinstance(d1[k], dict):
            _merge_dicts(d1[k], d2[k])
        else:
            d1[k] = v



import json
import logging
from pathlib import Path
from typing import Any, Dict, List

import cattr
import pytest

from typed_settings import _core
from typed_settings._dict_utils import _deep_options
from typed_settings.attrs import option, settings
from typed_settings.converters import default_converter, register_strlist_hook
from typed_settings.loaders import EnvLoader, FileLoader, Loader, TomlFormat


@settings(frozen=True)
class Host:
    name: str
    port: int


@settings(frozen=True)
class Settings:
    host: Host
    url: str
    default: int = 3


class TestLoadSettings:
    """Tests for load_settings()."""

    config = """[example]
        url = "https://example.com"
        [example.host]
        name = "example.com"
        port = 443
    """

    @pytest.fixture
    def config_files(self, tmp_path: Path) -> List[Path]:
        config_file = tmp_path.joinpath("settings.toml")
        config_file.write_text(self.config)
        return [config_file]

    @pytest.fixture
    def env_vars(self, monkeypatch: pytest.MonkeyPatch) -> None:
        monkeypatch.setenv("EXAMPLE_HOST_PORT", "42")

    @pytest.fixture
    def loaders(
        self, config_files: List[Path], env_vars: None
    ) -> List[Loader]:
        return [
            FileLoader(
                formats={"*.toml": TomlFormat("example")},
                files=config_files,
            ),
            EnvLoader(prefix="EXAMPLE_"),
        ]

    def test__load_settings(self, loaders: List[Loader]):
        """
        "_load_settings()" is the internal core loader and takes a list of
        options instead of a normal settings class.  It returns a dict and
        not a settings instance.
        """
        settings = _core._load_settings(
            cls=Settings,
            options=_deep_options(Settings),
            loaders=loaders,
        )
        assert settings == {
            "url": "https://example.com",
            "default": 3,  # This is from the cls
            "host": {
                "name": "example.com",
                "port": "42",  # Value not yet converted
            },
        }

    def test_load_settings(self, loaders: List[Loader]):
        """
        The "load_settings()" works like "_load_settings" but takes a settings
        class and returns an instance of it.
        """
        settings = _core.load_settings(
            cls=Settings,
            loaders=loaders,
        )
        assert settings == Settings(
            url="https://example.com",
            default=3,
            host=Host(
                name="example.com",
                port=42,
            ),
        )

    def test_load(self, config_files: List[Path], env_vars: None):
        """
        The "load()" shortcut automaticaly defines a file loader and an
        env loader.  Section and env var names are derived from the app name.
        """
        settings = _core.load(
            cls=Settings,
            appname="example",
            config_files=config_files,
        )
        assert settings == Settings(
            url="https://example.com",
            default=3,
            host=Host(
                name="example.com",
                port=42,  # Loaded from env var
            ),
        )

    def test_explicit_section(self, tmp_path: Path):
        """
        The automatically derived config section name name can be overriden.
        """
        config_file = tmp_path.joinpath("settings.toml")
        config_file.write_text(
            """[le-section]
            spam = "eggs"
        """
        )

        @settings(frozen=True)
        class Settings:
            spam: str = ""

        result = _core.load(
            cls=Settings,
            appname="example",
            config_files=[config_file],
            config_file_section="le-section",
        )
        assert result == Settings(spam="eggs")

    def test_explicit_files_var(
        self, tmp_path: Path, monkeypatch: pytest.MonkeyPatch
    ):
        """
        The automatically derived settings files var name can be overriden.
        """
        config_file = tmp_path.joinpath("settings.toml")
        config_file.write_text(
            """[example]
            spam = "eggs"
        """
        )

        monkeypatch.setenv("LE_SETTINGS", str(config_file))

        @settings(frozen=True)
        class Settings:
            spam: str = ""

        result = _core.load(
            cls=Settings,
            appname="example",
            config_files=[],
            config_files_var="LE_SETTINGS",
        )
        assert result == Settings(spam="eggs")

    def test_no_files_var(
        self, tmp_path: Path, monkeypatch: pytest.MonkeyPatch
    ):
        """
        Setting config files via an env var can be disabled.
        """
        config_file = tmp_path.joinpath("settings.toml")
        config_file.write_text(
            """[example]
            spam = "eggs"
        """
        )

        monkeypatch.setenv("EXAMPLE_SETTINGS", str(config_file))

        @settings(frozen=True)
        class Settings:
            spam: str = ""

        result = _core.load(
            cls=Settings,
            appname="example",
            config_files=[],
            config_files_var=None,
        )
        assert result == Settings(spam="")

    def test_env_var_dash_underscore(self, monkeypatch, tmp_path):
        """
        Dashes in the appname get replaced with underscores for the settings
        fiels var name.
        """

        @settings(frozen=True)
        class Settings:
            option: bool = True

        sf = tmp_path.joinpath("settings.toml")
        sf.write_text("[a-b]\noption = false\n")
        monkeypatch.setenv("A_B_SETTINGS", str(sf))

        result = _core.load(Settings, appname="a-b")
        assert result == Settings(option=False)

    def test_explicit_env_prefix(self, monkeypatch: pytest.MonkeyPatch):
        monkeypatch.setenv("P_SPAM", "spam")

        @settings(frozen=True)
        class Settings:
            spam: str = ""

        result = _core.load(
            cls=Settings, appname="example", config_files=[], env_prefix="P_"
        )
        assert result == Settings(spam="spam")

    def test_disable_env_vars(self, monkeypatch):
        monkeypatch.setenv("EXAMPLE_SPAM", "spam")

        @settings(frozen=True)
        class Settings:
            spam: str = ""

        result = _core.load(
            cls=Settings, appname="example", config_files=[], env_prefix=None
        )
        assert result == Settings(spam="")

    def test_load_nested_settings_by_default(self):
        """
        Instantiate nested settings with default settings and pass it to the
        parent settings even if no nested settings are defined in a config
        file or env var.

        Otherwise, the parent classed needed to set a default_factory for
        creating a nested settings instance.
        """

        @settings
        class Nested:
            a: int = 3
            b: str = "spam"

        @settings
        class Settings:
            nested: Nested

        s = _core.load(Settings, "test")
        assert s == Settings(Nested())

    def test_default_factories(self):
        """
        The default value "attr.Factory" is handle as if "attr.NOTHING" was
        set.

        See: https://gitlab.com/sscherfke/typed-settings/-/issues/6
        """

        @settings
        class S:
            opt: List[int] = option(factory=list)

        result = _core.load(S, "t")
        assert result == S()

    def test_custom_converter(self, monkeypatch: pytest.MonkeyPatch):
        """
        A custom cattr converter can be used in "load_settings()".
        """

        class Test:
            def __init__(self, x: int):
                self.attr = x

            def __eq__(self, other):
                return self.attr == other.attr

        @settings
        class Settings:
            opt: Test

        monkeypatch.setenv("TEST_OPT", "42")

        converter = cattr.GenConverter()
        converter.register_structure_hook(Test, lambda v, t: Test(int(v)))

        result = _core.load_settings(Settings, [EnvLoader("TEST_")], converter)
        assert result == Settings(Test(42))

    @pytest.mark.parametrize(
        "vals, kwargs",
        [
            (["3,4,42", "spam,eggs"], {"sep": ","}),
            (["3:4:42", "spam:eggs"], {"sep": ":"}),
            (['[3,4,"42"]', '["spam","eggs"]'], {"fn": json.loads}),
        ],
    )
    def test_load_list_from_env(
        self,
        vals: List[str],
        kwargs: Dict[str, Any],
        loaders: List[Loader],
        tmp_path: Path,
        monkeypatch: pytest.MonkeyPatch,
    ):
        """
        Lists can be loaded from env vars
        """
        c = default_converter()
        register_strlist_hook(c, **kwargs)

        @settings
        class Settings:
            x: List[int]
            y: List[Path]
            z: List[int]

        # The str2list hook should not interfere with loading of "normal"
        # lists.
        sf = tmp_path.joinpath("settings.toml")
        sf.write_text("[example]\nz = [1, 2]\n")

        monkeypatch.setenv("EXAMPLE_X", vals[0])
        monkeypatch.setenv("EXAMPLE_Y", vals[1])

        result = _core.load_settings(Settings, loaders, c)
        assert result == Settings(
            x=[3, 4, 42], y=[Path("spam"), Path("eggs")], z=[1, 2]
        )


class TestLogging:
    """
    Test emitted log messages.
    """

    def test_successfull_loading(self, caplog, tmp_path, monkeypatch):
        """
        In case of success, only DEBUG messages are emitted.
        """

        @settings
        class S:
            opt: str

        sf1 = tmp_path.joinpath("sf1.toml")
        sf1.write_text('[test]\nopt = "spam"\n')
        sf2 = tmp_path.joinpath("sf2.toml")
        sf2.write_text('[test]\nopt = "eggs"\n')
        monkeypatch.setenv("TEST_SETTINGS", str(sf2))
        monkeypatch.setenv("TEST_OPT", "bacon")

        caplog.set_level(logging.DEBUG)

        _core.load(S, "test", [sf1])

        assert caplog.record_tuples == [
            (
                "typed_settings",
                logging.DEBUG,
                "Env var for config files: TEST_SETTINGS",
            ),
            ("typed_settings", logging.DEBUG, f"Loading settings from: {sf1}"),
            ("typed_settings", logging.DEBUG, f"Loading settings from: {sf2}"),
            (
                "typed_settings",
                logging.DEBUG,
                "Looking for env vars with prefix: TEST_",
            ),
            ("typed_settings", logging.DEBUG, "Env var found: TEST_OPT"),
        ]

    def test_optional_files_not_found(self, caplog, tmp_path, monkeypatch):
        """
        Non-existing optional files emit an INFO message if file was specified
        by the app (passed to "load_settings()") an a WARNING message if the
        file was specified via an environment variable.
        """

        @settings
        class S:
            opt: str = ""

        sf1 = tmp_path.joinpath("sf1.toml")
        sf2 = tmp_path.joinpath("sf2.toml")
        monkeypatch.setenv("TEST_SETTINGS", str(sf2))

        caplog.set_level(logging.DEBUG)

        _core.load(S, "test", [sf1])

        assert caplog.record_tuples == [
            (
                "typed_settings",
                logging.DEBUG,
                "Env var for config files: TEST_SETTINGS",
            ),
            ("typed_settings", logging.INFO, f"Config file not found: {sf1}"),
            (
                "typed_settings",
                logging.WARNING,
                f"Config file from TEST_SETTINGS not found: {sf2}",
            ),
            (
                "typed_settings",
                logging.DEBUG,
                "Looking for env vars with prefix: TEST_",
            ),
            ("typed_settings", logging.DEBUG, "Env var not found: TEST_OPT"),
        ]

    def test_mandatory_files_not_found(self, caplog, tmp_path, monkeypatch):
        """
        In case of success, only ``debug`` messages are emitted.
        """

        @settings
        class S:
            opt: str = ""

        sf1 = tmp_path.joinpath("sf1.toml")
        monkeypatch.setenv("TEST_SETTINGS", f"!{sf1}")

        caplog.set_level(logging.DEBUG)

        with pytest.raises(FileNotFoundError):
            _core.load(S, "test")

        assert caplog.record_tuples == [
            (
                "typed_settings",
                logging.DEBUG,
                "Env var for config files: TEST_SETTINGS",
            ),
            (
                "typed_settings",
                logging.ERROR,
                f"Mandatory config file not found: {sf1}",
            ),
        ]



"""
Helpers for and additions to :mod:`attrs`.
"""
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Mapping,
    Optional,
    overload,
)

import attrs


if TYPE_CHECKING:
    try:
        from attr import (  # type: ignore
            _T,
            _ConverterType,
            _OnSetAttrArgType,
            _ReprArgType,
            _ValidatorArgType,
        )
    except ImportError:
        # Just in case the symbols are moved from "attr" to "attrs"
        from attrs import (  # type: ignore
            _T,
            _ConverterType,
            _OnSetAttrArgType,
            _ReprArgType,
            _ValidatorArgType,
        )

from .hooks import auto_convert


__all__ = [
    "METADATA_KEY",
    "SECRET",
    "auto_convert",
    "evolve",
    "option",
    "secret",
    "settings",
]


METADATA_KEY = "typed_settings"


class _SecretRepr:
    def __call__(self, _v) -> str:
        return "***"

    def __repr__(self) -> str:
        return "***"


SECRET = _SecretRepr()


settings = attrs.define
"""An alias to :func:`attrs.define()`"""


@overload
def option(
    *,
    default: None = ...,
    validator: None = ...,
    repr: "_ReprArgType" = ...,
    hash: Optional[bool] = ...,
    init: bool = ...,
    metadata: Optional[Dict[Any, Any]] = ...,
    converter: None = ...,
    factory: None = ...,
    kw_only: bool = ...,
    eq: Optional[bool] = ...,
    order: Optional[bool] = ...,
    on_setattr: Optional["_OnSetAttrArgType"] = ...,
    help: Optional[str] = ...,
) -> Any:
    ...


# This form catches an explicit None or no default and infers the type from the
# other arguments.
@overload
def option(
    *,
    default: None = ...,
    validator: "Optional[_ValidatorArgType[_T]]" = ...,
    repr: "_ReprArgType" = ...,
    hash: Optional[bool] = ...,
    init: bool = ...,
    metadata: Optional[Dict[Any, Any]] = ...,
    converter: Optional["_ConverterType"] = ...,
    factory: "Optional[Callable[[], _T]]" = ...,
    kw_only: bool = ...,
    eq: Optional[bool] = ...,
    order: Optional[bool] = ...,
    on_setattr: "Optional[_OnSetAttrArgType]" = ...,
    help: Optional[str] = ...,
) -> "_T":
    ...


# This form catches an explicit default argument.
@overload
def option(
    *,
    default: "_T",
    validator: "Optional[_ValidatorArgType[_T]]" = ...,
    repr: "_ReprArgType" = ...,
    hash: Optional[bool] = ...,
    init: bool = ...,
    metadata: Optional[Dict[Any, Any]] = ...,
    converter: "Optional[_ConverterType]" = ...,
    factory: "Optional[Callable[[], _T]]" = ...,
    kw_only: bool = ...,
    eq: Optional[bool] = ...,
    order: Optional[bool] = ...,
    on_setattr: "Optional[_OnSetAttrArgType]" = ...,
    help: Optional[str] = ...,
) -> "_T":
    ...


# This form covers type=non-Type: e.g. forward references (str), Any
@overload
def option(
    *,
    default: Optional["_T"] = ...,
    validator: "Optional[_ValidatorArgType[_T]]" = ...,
    repr: "_ReprArgType" = ...,
    hash: Optional[bool] = ...,
    init: bool = ...,
    metadata: Optional[Dict[Any, Any]] = ...,
    converter: "Optional[_ConverterType]" = ...,
    factory: "Optional[Callable[[], _T]]" = ...,
    kw_only: bool = ...,
    eq: Optional[bool] = ...,
    order: Optional[bool] = ...,
    on_setattr: "Optional[_OnSetAttrArgType]" = ...,
    help: Optional[str] = ...,
) -> Any:
    ...


def option(
    *,
    default=attrs.NOTHING,
    validator=None,
    repr=True,
    hash=None,
    init=True,
    metadata=None,
    converter=None,
    factory=None,
    kw_only=False,
    eq=None,
    order=None,
    on_setattr=None,
    help=None,
):
    """An alias to :func:`attrs.field()`"""
    if help is not None:
        if metadata is None:
            metadata = {}
        metadata.setdefault(METADATA_KEY, {})["help"] = help

    return attrs.field(
        default=default,
        validator=validator,
        repr=repr,
        hash=hash,
        init=init,
        metadata=metadata,
        converter=converter,
        factory=factory,
        kw_only=kw_only,
        eq=eq,
        order=order,
        on_setattr=on_setattr,
    )


@overload
def secret(
    *,
    default: None = ...,
    validator: None = ...,
    repr: _SecretRepr = ...,
    hash: Optional[bool] = ...,
    init: bool = ...,
    metadata: Optional[Dict[Any, Any]] = ...,
    converter: None = ...,
    factory: None = ...,
    kw_only: bool = ...,
    eq: Optional[bool] = ...,
    order: Optional[bool] = ...,
    on_setattr: "Optional[_OnSetAttrArgType]" = ...,
    help: Optional[str] = ...,
) -> Any:
    ...


# This form catches an explicit None or no default and infers the type from the
# other arguments.
@overload
def secret(
    *,
    default: None = ...,
    validator: "Optional[_ValidatorArgType[_T]]" = ...,
    repr: _SecretRepr = ...,
    hash: Optional[bool] = ...,
    init: bool = ...,
    metadata: Optional[Dict[Any, Any]] = ...,
    converter: "Optional[_ConverterType]" = ...,
    factory: "Optional[Callable[[], _T]]" = ...,
    kw_only: bool = ...,
    eq: Optional[bool] = ...,
    order: Optional[bool] = ...,
    on_setattr: "Optional[_OnSetAttrArgType]" = ...,
    help: Optional[str] = ...,
) -> "_T":
    ...


# This form catches an explicit default argument.
@overload
def secret(
    *,
    default: "_T",
    validator: "Optional[_ValidatorArgType[_T]]" = ...,
    repr: _SecretRepr = ...,
    hash: Optional[bool] = ...,
    init: bool = ...,
    metadata: Optional[Dict[Any, Any]] = ...,
    converter: "Optional[_ConverterType]" = ...,
    factory: "Optional[Callable[[], _T]]" = ...,
    kw_only: bool = ...,
    eq: Optional[bool] = ...,
    order: Optional[bool] = ...,
    on_setattr: "Optional[_OnSetAttrArgType]" = ...,
    help: Optional[str] = ...,
) -> "_T":
    ...


# This form covers type=non-Type: e.g. forward references (str), Any
@overload
def secret(
    *,
    default: "Optional[_T]" = ...,
    validator: "Optional[_ValidatorArgType[_T]]" = ...,
    repr: _SecretRepr = ...,
    hash: Optional[bool] = ...,
    init: bool = ...,
    metadata: Optional[Dict[Any, Any]] = ...,
    converter: "Optional[_ConverterType]" = ...,
    factory: "Optional[Callable[[], _T]]" = ...,
    kw_only: bool = ...,
    eq: Optional[bool] = ...,
    order: Optional[bool] = ...,
    on_setattr: "Optional[_OnSetAttrArgType]" = ...,
    help: Optional[str] = ...,
) -> Any:
    ...


def secret(
    *,
    default=attrs.NOTHING,
    validator=None,
    repr=SECRET,
    hash=None,
    init=True,
    metadata=None,
    converter=None,
    factory=None,
    kw_only=False,
    eq=None,
    order=None,
    on_setattr=None,
    help=None,
):
    """
    An alias to :func:`option()` but with a default repr that hides screts.

    When printing a settings instances, secret settings will represented with
    `***` istead of their actual value.

    See also:

        All arguments are describted here:

        - :func:`option()`
        - :func:`attrs.field()`

    Example:

    .. code-block:: python

        >>> from typed_settings import settings, secret
        >>>
        >>> @settings
        ... class Settings:
        ...     password: str = secret()
        ...
        >>> Settings(password="1234")
        Settings(password=***)
    """
    if help is not None:
        if metadata is None:
            metadata = {}
        metadata.setdefault(METADATA_KEY, {})["help"] = help

    return attrs.field(
        default=default,
        validator=validator,
        repr=repr,
        hash=hash,
        init=init,
        metadata=metadata,
        converter=converter,
        factory=factory,
        kw_only=kw_only,
        eq=eq,
        order=order,
        on_setattr=on_setattr,
    )


def evolve(inst, **changes):
    """
    Create a new instance, based on *inst* with *changes* applied.

    If the old value of an attribute is an ``attrs`` class and the new value
    is a dict, the old value is updated recursively.

    .. warning::

       This function is very similar to :func:`attrs.evolve()`, but the
       ``attrs`` version is not updating values recursively.  Instead, it will
       just replace ``attrs`` instances with a dict.

    :param inst: Instance of a class with ``attrs`` attributes.
    :param changes: Keyword changes in the new copy.

    :return: A copy of inst with *changes* incorporated.

    :raise TypeError: If *attr_name* couldn't be found in the class
        ``__init__``.
    :raise attr.exceptions.NotAnAttrsClassError: If *cls* is not an ``attrs``
        class.

    ..  versionadded:: 1.0.0
    """
    cls = inst.__class__
    attribs = attrs.fields(cls)
    for a in attribs:
        if not a.init:
            continue
        attr_name = a.name  # To deal with private attributes.
        init_name = attr_name if attr_name[0] != "_" else attr_name[1:]
        old_value = getattr(inst, attr_name)
        if init_name not in changes:
            # Add original value to changes
            changes[init_name] = old_value
        elif attrs.has(old_value) and isinstance(changes[init_name], Mapping):
            # Evolve nested attrs classes
            changes[init_name] = evolve(old_value, **changes[init_name])

    return cls(**changes)



"""
Converters and helpers for :mod:`cattr`.
"""
from datetime import datetime
from enum import Enum
from pathlib import Path
from typing import Any, Callable, Optional, Type, Union

from attrs import has
from cattrs import Converter, GenConverter


try:
    from cattr._compat import (  # type: ignore
        is_frozenset,
        is_mutable_set,
        is_sequence,
        is_tuple,
    )
except ImportError:
    from cattrs._compat import (  # type: ignore
        is_frozenset,
        is_mutable_set,
        is_sequence,
        is_tuple,
    )

from .exceptions import InvalidValueError
from .types import ET, SettingsDict, T


def default_converter() -> GenConverter:
    """
    Get an instanceof the default converter used by Typed Settings.

    Return:
        A :class:`cattrs.GenConverter` configured with addional hooks for
        loading the follwing types:

        - :class:`bool` using :func:`.to_bool()`
        - :class:`datetime.datetime` using :func:`.to_dt()`
        - :class:`enum.Enum` using :func:`.to_enum()`
        - :class:`pathlib.Path`

        The converter can also structure attrs instances from existing attrs
        instances (normaly, it would only work with dicts).  This allows using
        instances of nested class es of default values for options.  See
        :func:`register_structure_hook_factory()`.

    This converter can also be used as a base for converters with custom
    structure hooks.
    """
    converter = GenConverter()
    register_attrs_hook_factory(converter)
    register_strlist_hook(converter, ":")
    for t, h in DEFAULT_STRUCTURE_HOOKS:
        converter.register_structure_hook(t, h)  # type: ignore
    return converter


def register_attrs_hook_factory(converter: Converter) -> None:
    """
    Register a hook factory that allows using instances of attrs classes where
    cattrs would normally expect a dictionary.

    These instances are then returned as-is and without further processing.
    """

    def allow_attrs_instances(typ):
        def structure_attrs(val, _):
            if isinstance(val, typ):
                return val
            return converter.structure_attrs_fromdict(val, typ)

        return structure_attrs

    converter.register_structure_hook_factory(has, allow_attrs_instances)


def register_strlist_hook(
    converter: Converter,
    sep: Optional[str] = None,
    fn: Optional[Callable[[str], list]] = None,
) -> None:
    """
    Register a hook factory with *converter* that allows structuring lists,
    (frozen) sets and tuples from strings (which may, e.g., come from
    environment variables).

    Args:
        converter: The converter to register the hooks with.
        sep: A separator used for splitting strings (see :meth:`str.split()`).
            Cannot be used together with *fn*.
        fn: A function that takes a string and returns a list, e.g.,
            :func:`json.loads()`.  Cannot be used together with *spe*.

    Example:

        .. code-block:: python

            >>> from typing import List
            >>>
            >>> converter = default_converter()
            >>> register_strlist_hook(converter, sep=":")
            >>> converter.structure("1:2:3", List[int])
            [1, 2, 3]
            >>>
            >>> import json
            >>>
            >>> converter = default_converter()
            >>> register_strlist_hook(converter, fn=json.loads)
            >>> converter.structure("[1,2,3]", List[int])
            [1, 2, 3]


    """
    if (sep is None and fn is None) or (sep is not None and fn is not None):
        raise ValueError('You may either pass "sep" *or* "fn"')
    if sep is not None:
        fn = lambda v: v.split(sep)  # noqa

    collection_types = [
        # Order is important, tuple must be last!
        (is_sequence, converter._structure_list),
        (is_mutable_set, converter._structure_set),
        (is_frozenset, converter._structure_frozenset),
        (is_tuple, converter._structure_tuple),
    ]

    for check, structure_func in collection_types:
        hook_factory = _generate_hook_factory(structure_func, fn)
        converter.register_structure_hook_factory(check, hook_factory)


def _generate_hook_factory(structure_func, fn):
    def gen_func(typ):
        def str2collection(val, _):
            if isinstance(val, str):
                val = fn(val)
            return structure_func(val, typ)

        return str2collection

    return gen_func


def from_dict(settings: SettingsDict, cls: Type[T], converter: Converter) -> T:
    """
    Convert a settings dict to an attrs class instance using a cattrs
    converter.

    Args:
        settings: Dictionary with settings
        cls: Attrs class to which the settings are converted to
        converter: Cattrs convert to use for the conversion

    Return:
        An instance of *cls*.

    Raise:
        InvalidValueError: If a value cannot be converted to the correct type.
    """
    try:
        return converter.structure_attrs_fromdict(settings, cls)
    except (AttributeError, KeyError, ValueError, TypeError) as e:
        raise InvalidValueError(str(e)) from e


def to_dt(value: Union[datetime, str], _type: type = datetime) -> datetime:
    """
    Convert an ISO formatted string to :class:`datetime.datetime`.  Leave the
    input untouched if it is already a datetime.

    See: :func:`datetime.datetime.fromisoformat()`

    The ``Z`` suffix is also supported and will be replaced with ``+00:00``.

    Args:
        value: The input data
        _type: The desired output type, will be ignored

    Return:
        The converted datetime instance

    Raise:
        TypeError: If *val* is neither a string nor a datetime
    """
    if not isinstance(value, (datetime, str)):
        raise TypeError(
            f'Invalid type "{type(value).__name__}"; expected "datetime" or '
            f'"str".'
        )
    if isinstance(value, str):
        if value[-1] == "Z":
            value = value.replace("Z", "+00:00")
        return datetime.fromisoformat(value)
    return value


def to_bool(value: Any, _type: type = bool) -> bool:
    """
    Convert "boolean" strings (e.g., from env. vars.) to real booleans.

    Values mapping to :code:`True`:

    - :code:`True`
    - :code:`"true"` / :code:`"t"`
    - :code:`"yes"` / :code:`"y"`
    - :code:`"on"`
    - :code:`"1"`
    - :code:`1`

    Values mapping to :code:`False`:

    - :code:`False`
    - :code:`"false"` / :code:`"f"`
    - :code:`"no"` / :code:`"n"`
    - :code:`"off"`
    - :code:`"0"`
    - :code:`0`

    Raise :exc:`ValueError` for any other value.
    """
    if isinstance(value, str):
        value = value.lower()
    truthy = {True, "true", "t", "yes", "y", "on", "1", 1}
    falsy = {False, "false", "f", "no", "n", "off", "0", 0}
    try:
        if value in truthy:
            return True
        if value in falsy:
            return False
    except TypeError:
        # Raised when "val" is not hashable (e.g., lists)
        pass
    raise ValueError(f"Cannot convert value to bool: {value}")


def to_enum(value: Any, cls: Type[ET]) -> ET:
    """
    Return a converter that creates an instance of the :class:`.Enum` *cls*.

    If the to be converted value is not already an enum, the converter will
    create one by name (``MyEnum[val]``).

    Args:
        value: The input data
        cls: The enum type

    Return:
        An instance of *cls*

    Raise:
        KeyError: If *value* is not a valid member of *cls*

    """
    if isinstance(value, cls):
        return value

    return cls[value]


def to_path(value: Union[Path, str], _type: type) -> Path:
    return Path(value)


DEFAULT_STRUCTURE_HOOKS = [
    (bool, to_bool),
    (datetime, to_dt),
    (Enum, to_enum),
    (Path, to_path),
]



import importlib.util
import logging
import os
from fnmatch import fnmatch
from pathlib import Path
from typing import Any, Callable, Dict, Iterable, List, Optional, Union


try:
    from typing import Protocol
except ImportError:
    # Python 3.7
    from typing import _Protocol as Protocol  # type: ignore

import attr
import toml

from ._dict_utils import _merge_dicts, _set_path
from .exceptions import (
    ConfigFileLoadError,
    ConfigFileNotFoundError,
    InvalidOptionsError,
    UnknownFormatError,
)
from .types import OptionList, SettingsDict


LOGGER = logging.getLogger("typed_settings")


class Loader(Protocol):
    """
    **Protocol:** Methods that settings loaders must implement.

    Custom settings loaders must implement this.

    .. versionchanged:: 1.0.0
       Renamed ``load()`` to ``__call__()`` and also pass the settings
       class.
    """

    def __call__(
        self, settings_cls: type, options: OptionList
    ) -> SettingsDict:
        """
        Load settings for the given options.

        Args:
            options: The list of available settings.
            settings_cls: The base settings class for all options.

        Return:
            A dict with the loaded settings.
        """
        ...


class FileFormat(Protocol):
    """
    **Protocol:** Methods that file format loaders for :class:`FileLoader`
    must implement.

    Custom file format loaders must implement this.

    .. versionchanged:: 1.0.0
       Renamed ``load_file()`` to ``__call__()`` and also pass the settings
       class.
    """

    def __call__(
        self, path: Path, settings_cls: type, options: OptionList
    ) -> SettingsDict:
        """
        Load settings from a given file and return them as a dict.

        Args:
            path: The path to the config file.
            options: The list of available settings.
            settings_cls: The base settings class for all options.

        Return:
            A dict with the loaded settings.

        Raise:
            ConfigFileNotFoundError: If *path* does not exist.
            ConfigFileLoadError: If *path* cannot be read/loaded/decoded.
        """
        ...


class InstanceLoader:
    """
    Load settings from an instance of the settings class.

    Args:
        instance: The settings instance from which to load option values.

    .. versionadded:: 1.0.0
    """

    def __init__(self, instance: object):
        self.instance = instance

    def __call__(
        self, settings_cls: type, options: OptionList
    ) -> SettingsDict:
        """
        Load settings for the given options.

        Args:
            options: The list of available settings.
            settings_cls: The base settings class for all options.

        Return:
            A dict with the loaded settings.
        """
        if not isinstance(self.instance, settings_cls):
            raise ValueError(
                f'"self.instance" is not an instance of {settings_cls}: '
                f"{type(self.instance)}"
            )
        return attr.asdict(self.instance)


class EnvLoader:
    """
    Load settings from environment variables.

    Args:
        prefix: Prefix for environment variables, e.g., ``MYAPP_``.
    """

    def __init__(self, prefix: str):
        self.prefix = prefix

    def __call__(
        self, settings_cls: type, options: OptionList
    ) -> SettingsDict:
        """
        Load settings for the given options.

        Args:
            options: The list of available settings.
            settings_cls: The base settings class for all options.

        Return:
            A dict with the loaded settings.
        """
        prefix = self.prefix
        LOGGER.debug(f"Looking for env vars with prefix: {prefix}")

        env = os.environ
        values: Dict[str, Any] = {}
        for o in options:
            varname = f"{prefix}{o.path.upper().replace('.', '_')}"
            if varname in env:
                LOGGER.debug(f"Env var found: {varname}")
                _set_path(values, o.path, env[varname])
            else:
                LOGGER.debug(f"Env var not found: {varname}")

        return values


class FileLoader:
    """
    Load settings from config files.

    Settings of multiple files will be merged.  The last file has the highest
    precedence.  Files specified via an environment variable are loaded after
    the files passed to this class, i.e.:

    - First file from *files*
    - ...
    - Last file from *files*
    - First file from *env_var*
    - ...
    - Last file from *env_var*

    Mandatory files can be prefixed with ``!``.  Optional files will be ignored
    if they don't exist.

    Args:
        formats: A dict mapping glob patterns to :class:`FileFormat` instances.
        files: A list of filenames to try to load.
        env_var: Name of the environment variable that may hold additional file
            paths.  If it is ``None``, only files from *files* will be loaded.
    """

    def __init__(
        self,
        formats: Dict[str, FileFormat],
        files: Iterable[Union[str, Path]],
        env_var: Optional[str] = None,
    ):
        self.files = files
        self.env_var = env_var
        self.formats = formats

    def __call__(
        self, settings_cls: type, options: OptionList
    ) -> SettingsDict:
        """
        Load settings for the given options.

        Args:
            options: The list of available settings.
            settings_cls: The base settings class for all options.

        Return:
            A dict with the loaded settings.

        Raise:
            UnknownFormat: When no :class:`FileFormat` is configured for a
                loaded file.
            ConfigFileNotFoundError: If *path* does not exist.
            ConfigFileLoadError: If *path* cannot be read/loaded/decoded.
            InvalidOptionError: If invalid settings have been found.
        """
        paths = self._get_config_filenames(self.files, self.env_var)
        merged_settings: SettingsDict = {}
        for path in paths:
            settings = self._load_file(path, settings_cls, options)
            _merge_dicts(merged_settings, settings)
        return merged_settings

    def _load_file(
        self,
        path: Path,
        settings_cls: type,
        options: OptionList,
    ) -> SettingsDict:
        """
        Load a file and return its cleaned contents
        """
        # "clean_settings()" must be called for each loaded file individually
        # because of the "-"/"_" normalization.  This also allows us to tell
        # the user the exact file that contains errors.
        for pattern, ffloader in self.formats.items():
            if fnmatch(path.name, pattern):
                settings = ffloader(path, settings_cls, options)
                settings = clean_settings(settings, options, path)
                return settings

        raise UnknownFormatError(f"No loader configured for: {path}")

    @staticmethod
    def _get_config_filenames(
        files: Iterable[Union[str, Path]], env_var: Optional[str]
    ) -> List[Path]:
        """
        Concatenate *config_files* and files from env var *config_files_var*.
        """
        candidates = [(False, str(f)) for f in files]
        if env_var:
            LOGGER.debug(f"Env var for config files: {env_var}")
            candidates += [
                (True, fname) for fname in os.getenv(env_var, "").split(":")
            ]
        else:
            LOGGER.debug("Env var for config files not set")

        paths = []
        for from_envvar, fname in candidates:
            _, flag, fname = fname.rpartition("!")
            if not fname:
                continue
            is_mandatory = flag == "!"
            try:
                path = Path(fname).resolve(strict=True)
            except FileNotFoundError:
                if is_mandatory:
                    LOGGER.error(f"Mandatory config file not found: {fname}")
                    raise
                if from_envvar:
                    LOGGER.warning(
                        f"Config file from {env_var} not found: {fname}"
                    )
                else:
                    LOGGER.info(f"Config file not found: {fname}")
            else:
                LOGGER.debug(f"Loading settings from: {path}")
                paths.append(Path(fname))

        return paths


class PythonFormat:
    """
    Support for Python files.  Read settings from the given *section*.

    Args:
        section: The config file section to load settings from.
    """

    def __init__(
        self,
        cls_name: str,
        key_transformer: Callable[[str], str] = lambda k: k,
        flat: bool = False,
    ):
        self.cls_name = cls_name
        self.key_transformer = key_transformer
        self.flat = flat

    @staticmethod
    def to_lower(text: str) -> str:
        """
        Transform *text* to lower case.
        """
        return text.lower()

    def __call__(
        self, path: Path, settings_cls: type, options: OptionList
    ) -> SettingsDict:
        """
        Load settings from a Python file and return them as a dict.

        Args:
            path: The path to the config file.
            options: The list of available settings.
            settings_cls: The base settings class for all options.

        Return:
            A dict with the loaded settings.

        Raise:
            ConfigFileNotFoundError: If *path* does not exist.
            ConfigFileLoadError: If *path* cannot be read/loaded/decoded.
        """
        module = self._import_module(path)

        def cls2dict(cls) -> SettingsDict:
            d = dict(cls.__dict__)
            result = {}
            for k, v in d.items():
                if k.startswith("_"):
                    continue
                k = self.key_transformer(k)
                if isinstance(v, type):
                    v = cls2dict(v)
                result[k] = v
            return result

        try:
            cls = getattr(module, self.cls_name)
        except AttributeError:
            return {}
        settings = cls2dict(cls)
        if self.flat:
            for o in options:
                key = f"{o.path.replace('.', '_')}"
                if key in settings:
                    val = settings.pop(key)
                    _set_path(settings, o.path, val)
        return settings

    def _import_module(self, path: Path) -> object:
        module_name = path.stem
        spec = importlib.util.spec_from_file_location(module_name, path)
        if spec is None or spec.loader is None:
            raise ConfigFileNotFoundError(
                "No such file or directory: '{path}'"
            )
        module = importlib.util.module_from_spec(spec)
        try:
            spec.loader.exec_module(module)  # type: ignore
        except SyntaxError as e:
            raise ConfigFileLoadError(str(e)) from e
        return module


class TomlFormat:
    """
    Support for TOML files.  Read settings from the given *section*.

    Args:
        section: The config file section to load settings from.
    """

    def __init__(self, section: str):
        self.section = section

    def __call__(
        self, path: Path, settings_cls: type, options: OptionList
    ) -> SettingsDict:
        """
        Load settings from a TOML file and return them as a dict.

        Args:
            path: The path to the config file.
            options: The list of available settings.
            settings_cls: The base settings class for all options.

        Return:
            A dict with the loaded settings.

        Raise:
            ConfigFileNotFoundError: If *path* does not exist.
            ConfigFileLoadError: If *path* cannot be read/loaded/decoded.
        """
        sections = self.section.split(".")
        try:
            settings = toml.load(path.open())
        except FileNotFoundError as e:
            raise ConfigFileNotFoundError(str(e)) from e
        except (PermissionError, toml.TomlDecodeError) as e:
            raise ConfigFileLoadError(str(e)) from e
        for s in sections:
            try:
                settings = settings[s]
            except KeyError:
                return {}
        return settings


def clean_settings(
    settings: SettingsDict, options: OptionList, source: Any
) -> SettingsDict:
    """
    Recursively check settings for invalid entries and raise an error.

    An error is not raised until all options have been checked.  It then lists
    all invalid options that have been found.

    Args:
        settings: The settings to be cleaned.
        options: The list of available settings.
        source: Source of the settings (e.g., path to a config file).
                It should have a useful string representation.

    Return:
        The cleaned settings.
    Raise:
        InvalidOptionError: If invalid settings have been found.
    """
    invalid_paths = []
    valid_paths = {o.path for o in options}
    cleaned: SettingsDict = {}

    def _iter_dict(d: SettingsDict, prefix: str) -> None:
        for key, val in d.items():
            key = key.replace("-", "_")
            path = f"{prefix}{key}"

            if path in valid_paths:
                _set_path(cleaned, path, val)
                continue

            if isinstance(val, dict):
                _iter_dict(val, f"{path}.")
            else:
                invalid_paths.append(path)

    _iter_dict(settings, "")

    if invalid_paths:
        joined_paths = ", ".join(sorted(invalid_paths))
        raise InvalidOptionsError(
            f"Invalid options found in {source}: {joined_paths}"
        )

    return cleaned



# pypirc_0.py

import sys
from typing import List

import typed_settings as ts


@ts.settings
class RepoServer:
    repository: str
    username: str
    password: str = ts.secret(default="")


@ts.settings
class Settings:
    index_servers: List[str]


REPO_NAME = sys.argv[1]
repo_server = ts.load(RepoServer, REPO_NAME, ["pypirc.toml"])
print(repo_server)



# python_gitlab.py

import typed_settings as ts


@ts.settings
class GitlabSettings:
    url: str
    private_token: str = ts.secret()
    api_version: int = 3


@ts.settings
class GlobalSettings:
    default: str
    ssl_verify: bool = True


global_settings = ts.load(
    GlobalSettings,
    appname="gitlab",
    config_files=["python-gitlab.toml"],
    config_file_section="global",
)
gitlab_settings = ts.load(
    GitlabSettings,
    appname="gitlab",
    config_files=["python-gitlab.toml"],
    config_file_section=global_settings.default,
)
print(global_settings)
print(gitlab_settings)
